/*-
 * Copyright 2018 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Nils Goroll <nils.goroll@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <cache/cache.h>
#include <vcl.h>

#include "vcc_frozen_if.h"
#include "vmod_frozen.h"
#include "frozen/frozen.h"

static enum type_e
type_parse(VCL_ENUM e) {
#define VMODENUM(n) if (e == VENUM(n)) return(type_ ## n);
#include "tbl_enum_type.h"
	WRONG("illegal type enum");
}

static const char const *type_s[_TYPE_E_MAX] = {
	[_TYPE_E_INVALID] = NULL,
#define VMODENUM(n) [type_ ## n] = #n,
#include "tbl_enum_type.h"
};

/*
 * for the vmod object, this struct is the specification
 *
 * .parse memcopies to array of expects and fills out the bottom
 */
struct vmod_frozen_expect {
	unsigned	magic;
#define VMOD_FROZEN_EXPECT_MAGIC	0x7d6dcdeb
	unsigned int	nullok:1;
	unsigned int	required:1;
	unsigned int	valid:1;	// written by .parse
	unsigned int	null:1;	// written by .parse
	const char	*path;

	// written by .parse
	const char	*val;
	int		len;
	// expected type initially, then overwritten
	enum type_e	type;
};

struct vmod_frozen_parser {
	unsigned			magic;
#define VMOD_FROZEN_PARSER_MAGIC	0x3c438bd9

	int				limit;
	char				*vcl_name;
	int				n_expect;
	int				s_expect;
	struct vmod_frozen_expect	*expects;
};

struct vmod_frozen_task {
	unsigned			magic;
#define VMOD_FROZEN_TASK_MAGIC	0x5028fbb3
	int				n_expect;
	int				n_seen;
	int				first;
	struct vmod_frozen_expect	*vals;
};

VCL_VOID
vmod_parser__init(VRT_CTX,
    struct vmod_frozen_parser **vfpap, const char *vcl_name, VCL_INT limit)
{
	struct vmod_frozen_parser *vfpa;
	struct vmod_frozen_expect *vfex;
	const int min_s_expect = 64 / sizeof(*vfpa);

	AN(vfpap);
	AZ(*vfpap);

	AN(min_s_expect);

	if (limit < 1) {
		VRT_fail(ctx, "frozen.parser limit must be 1 or higher");
		return;
	}

	ALLOC_OBJ(vfpa, VMOD_FROZEN_PARSER_MAGIC);
	if (vfpa == NULL) {
		VRT_fail(ctx, "frozen.parser obj alloc failed");
		return;
	}

	REPLACE(vfpa->vcl_name, vcl_name);
	if (vfpa->vcl_name == NULL) {
		VRT_fail(ctx, "frozen.parser dup vcl_name failed");
		goto err_dup;
	}

	vfex = malloc(sizeof(*vfex) * min_s_expect);
	if (vfex == NULL) {
		VRT_fail(ctx, "frozen.parser alloc expects failed");
		goto err_exp;
	}

	vfpa->limit = limit;
	vfpa->s_expect = min_s_expect;
	vfpa->expects = vfex;

	*vfpap = vfpa;

	return;

  err_exp:
	free(vfpa->vcl_name);
  err_dup:
	FREE_OBJ(vfpa);
}

VCL_VOID
vmod_parser__fini(struct vmod_frozen_parser **vfpap)
{
	struct vmod_frozen_parser *vfpa = *vfpap;
	struct vmod_frozen_expect *vfex;
	int i;

	*vfpap = NULL;
	if (vfpa == NULL)
		return;

	CHECK_OBJ(vfpa, VMOD_FROZEN_PARSER_MAGIC);

	for (i = 0; i < vfpa->n_expect; i++) {
		vfex = &vfpa->expects[i];
		free(TRUST_ME(vfex->path));
	}

	free(vfpa->expects);
	free(vfpa->vcl_name);
	FREE_OBJ(vfpa);
}

VCL_VOID
vmod_parser_expect(VRT_CTX,
    struct vmod_frozen_parser *vfpa, VCL_STRING path, VCL_ENUM type_s,
    VCL_BOOL null, VCL_BOOL required)
{
	struct vmod_frozen_expect *vfex;
	enum type_e type = type_parse(type_s);
	struct vmod_priv *task;
	int n;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(vfpa, VMOD_FROZEN_PARSER_MAGIC);

	if (ctx->method != VCL_MET_INIT) {
		VRT_fail(ctx, "%s.expect() may only be called from vcl_init {}",
		    vfpa->vcl_name);
		return;
	}

	if (vfpa->n_expect == vfpa->s_expect) {
		n = vfpa->s_expect << 1;
		vfex = realloc(vfpa->expects, n * sizeof(*vfex));
		if (vfex == NULL) {
			VRT_fail(ctx, "%s.expect() realloc %d elements failed",
			    vfpa->vcl_name, n);
			return;
		}
		vfpa->expects = vfex;
		vfpa->s_expect = n;
	}

	/* invalidate parse task in case expect is called after parse */
	task = VRT_priv_task(ctx, vfpa);
	if (task != NULL)
		task->priv = NULL;

	assert(vfpa->s_expect > vfpa->n_expect);
	vfex = &vfpa->expects[vfpa->n_expect++];

	INIT_OBJ(vfex, VMOD_FROZEN_EXPECT_MAGIC);
	vfex->path = strdup(path);
	vfex->type = type;
	vfex->nullok = !!null;
	vfex->required = !!required;
}

static void vfex_cb(void *, const char *, size_t, const char *,
    const struct json_token *);

VCL_BOOL
vmod_parser_parse(VRT_CTX, struct vmod_frozen_parser *vfpaa,
    const char *s)
{
	const struct vmod_frozen_parser *vfpa = vfpaa;
	struct vmod_frozen_task *vfta;
	struct vmod_frozen_expect *vfex;
	struct vmod_priv *task;
	struct frozen_args args[1];
	VCL_BOOL r;
	int i;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(vfpa, VMOD_FROZEN_PARSER_MAGIC);

	INIT_FROZEN_ARGS(args);
	args->limit = vfpa->limit;

	if (vfpa->n_expect == 0)
		return (!! json_walk_args(s, strlen(s), args));

	assert(vfpa->n_expect > 0);

	task = VRT_priv_task(ctx, vfpa);

	if (task == NULL) {
		VRT_fail(ctx, "no priv_task");
		return (0);
	}

	if (task->priv) {
		CAST_OBJ_NOTNULL(vfta, task->priv,
		    VMOD_FROZEN_TASK_MAGIC);
		vfex = vfta->vals;
		INIT_OBJ(vfta, VMOD_FROZEN_TASK_MAGIC);
	} else {
		vfta = WS_Alloc(ctx->ws, sizeof *vfta);
		if (vfta == NULL) {
			VRT_fail(ctx, "%s.parse() WS_Alloc task failed",
			    vfpa->vcl_name);
			return (0);
		}
		INIT_OBJ(vfta, VMOD_FROZEN_TASK_MAGIC);

		vfex = WS_Alloc(ctx->ws, vfpa->n_expect * sizeof *vfex);
		if (vfex == NULL) {
			VRT_fail(ctx, "%s.parse() WS_Alloc task failed",
			    vfpa->vcl_name);
			return (0);
		}
		task->priv = vfta;
	}
	AN(vfex);
	vfta->n_expect = vfpa->n_expect;
	vfta->vals = vfex;
	memcpy(vfex, vfpa->expects, vfpa->n_expect * sizeof *vfex);

	args->callback = vfex_cb;
	args->callback_data = vfta;

	r = !! json_walk_args(s, strlen(s), args);

	// prep for extract
	vfta->first = 0;

	if (r == 0)
		return (r);
	if (vfta->n_expect == vfta->n_seen)
		return (r);
	for (i = 0; i < vfta->n_expect; i++) {
		vfex = &vfta->vals[i];
		if (vfex->required == 1 &&
		    vfex->valid == 0)
			return (0);
	}
	return(r);
}

static const enum type_e json_type_2_type_e[JSON_TYPES_CNT] = {
	[JSON_TYPE_INVALID] = _TYPE_E_INVALID,
	[JSON_TYPE_STRING] = type_STRING,
	[JSON_TYPE_NUMBER] = type_NUMBER,
	[JSON_TYPE_TRUE] = type_BOOL,
	[JSON_TYPE_FALSE] = type_BOOL,
	[JSON_TYPE_NULL] = _TYPE_E_INVALID,
	[JSON_TYPE_OBJECT_START] = _TYPE_E_INVALID,
	[JSON_TYPE_OBJECT_END] = type_OBJECT,
	[JSON_TYPE_ARRAY_START] = _TYPE_E_INVALID,
	[JSON_TYPE_ARRAY_END] = type_ARRAY
};

static void
vfex_cb(void *callback_data, const char *name,
   size_t name_len, const char *path, const struct json_token *token)
{
	struct vmod_frozen_task *vfta;
	struct vmod_frozen_expect *vfex;
	enum type_e type;
	int i;

	CAST_OBJ_NOTNULL(vfta, callback_data,
	    VMOD_FROZEN_TASK_MAGIC);

	(void) name;
	(void) name_len;

	if (vfta->n_expect == vfta->n_seen)
		return;

	AN(token);
	assert(token->type != JSON_TYPE_INVALID);
	if (token->type == JSON_TYPE_OBJECT_START ||
	    token->type == JSON_TYPE_ARRAY_START)
		return;

	assert(vfta->n_expect > vfta->n_seen);

	/* vfta->first marks the first expect not yet valid */
	for (i = vfta->first; i < vfta->n_expect; i++) {
		vfex = &vfta->vals[i];

		if (vfex->valid)
			continue;

		if (token->type == JSON_TYPE_NULL &&
		    vfex->nullok == 0)
			continue;

		type = json_type_2_type_e[token->type];

		if (vfex->type != type_ANY &&
		    vfex->type != type)
			continue;

		if (strcmp(vfex->path, path))
			continue;

		vfex->valid = 1;
		vfex->val = token->ptr;
		vfex->len = token->len;
		if (token->type == JSON_TYPE_NULL)
			vfex->null = 1;
		else
			vfex->type = type;

		vfta->n_seen++;

		/* advance vfta->first only if this is the first vfex */
		if (i != vfta->first)
			return;

		while (++i < vfta->n_expect) {
			vfex = &vfta->vals[i];
			if (vfex->valid == 0) {
				vfta->first = i;
				return;
			}
		}
		vfta->first = i;
		assert (vfta->n_seen == vfta->n_expect);
		return;
	}
}

VCL_STRING
vmod_parser_extract(VRT_CTX,
    struct vmod_frozen_parser *vfpaa, VCL_STRING path, VCL_STRING r_null,
    VCL_STRING r_undef)
{
	const struct vmod_frozen_parser *vfpa = vfpaa;
	struct vmod_frozen_task *vfta;
	struct vmod_frozen_expect *vfex;
	struct vmod_priv *task;
	int i, n;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(vfpa, VMOD_FROZEN_PARSER_MAGIC);

	task = VRT_priv_task(ctx, vfpa);

	if (task == NULL) {
		VRT_fail(ctx, "no priv_task");
		return (r_undef);
	}

	if (task->priv == NULL) {
		VRT_fail(ctx, "no state from .parse()");
		return (r_undef);
	}

	CAST_OBJ_NOTNULL(vfta, task->priv,
	    VMOD_FROZEN_TASK_MAGIC);

	if (vfta->n_seen == 0)
		return (r_undef);

	for (n = 0; n < vfta->n_expect; n++) {
		i = (vfta->first + n) % vfta->n_expect;

		vfex = &vfta->vals[i];

		if (strcmp(vfex->path, path))
			continue;

		if (vfex->valid == 0)
			return (r_undef);

		if (vfex->null == 1)
			return (r_null);

		vfta->first = i;

		if (vfex->val == NULL || vfex->len == 0)
			return (r_undef);

		return (WS_Printf(ctx->ws, "%.*s", vfex->len, vfex->val));
	}

	return (r_undef);
}

VCL_STRING vmod_parser_type(VRT_CTX,
    struct vmod_frozen_parser *vfpaa, VCL_STRING path)
{
	const struct vmod_frozen_parser *vfpa = vfpaa;
	struct vmod_frozen_task *vfta;
	struct vmod_frozen_expect *vfex;
	struct vmod_priv *task;
	int i, n;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(vfpa, VMOD_FROZEN_PARSER_MAGIC);

	task = VRT_priv_task(ctx, vfpa);

	if (task == NULL) {
		VRT_fail(ctx, "no priv_task");
		return (NULL);
	}

	if (task->priv == NULL) {
		VRT_fail(ctx, "no state from .match()");
		return (NULL);
	}

	CAST_OBJ_NOTNULL(vfta, task->priv,
	    VMOD_FROZEN_TASK_MAGIC);

	if (vfta->n_seen == 0)
		return (NULL);

	for (n = 0; n < vfta->n_expect; n++) {
		i = (vfta->first + n) % vfta->n_expect;
		vfex = &vfta->vals[i];

		if (strcmp(vfex->path, path))
			continue;

		if (vfex->valid == 0)
			return (NULL);

		if (vfex->null == 1)
			return (NULL);

		assert(vfex->type != type_ANY);

		return (type_s[vfex->type]);
	}

	return (NULL);
}
